from __future__ import unicode_literals

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

if hasattr(settings, 'PAYFAST_URL_BASE'):
    URL_BASE = settings.PAYFAST_URL_BASE
else:
    raise ImproperlyConfigured(
        'Please configure settings.PAYFAST_URL_BASE with the base URL of your site'
        ' (as a string, or a callable returning a string)')

TEST_MODE = getattr(settings, 'PAYFAST_TEST_MODE', False)

TEST_MERCHANT_ID = '10000100'
TEST_MERCHANT_KEY = '46f0cd694581a'

if TEST_MODE:
    # real id and key don't work in sandbox
    MERCHANT_ID = TEST_MERCHANT_ID
    MERCHANT_KEY = TEST_MERCHANT_KEY
else:
    MERCHANT_ID = getattr(settings, 'PAYFAST_MERCHANT_ID')
    MERCHANT_KEY = getattr(settings, 'PAYFAST_MERCHANT_KEY')

LIVE_SERVER = 'https://www.payfast.co.za'
SANDBOX_SERVER = 'https://sandbox.payfast.co.za'
SERVER = SANDBOX_SERVER if TEST_MODE else LIVE_SERVER

PROCESS_URL = SERVER + '/eng/process'

REQUIRE_AMOUNT_MATCH = getattr(settings, 'PAYFAST_REQUIRE_AMOUNT_MATCH', True)
USE_POSTBACK = getattr(settings, 'PAYFAST_USE_POSTBACK', True)

# request.META key with client ip address
IP_HEADER = getattr(settings, 'PAYFAST_IP_HEADER', 'REMOTE_ADDR')

# Reference: https://developers.payfast.co.za/documentation/#ip-addresses
# The values below are current as of 2017 December.
DEFAULT_PAYFAST_IP_ADDRESSES = [
    '197.97.145.144/28',
    '41.74.179.192/27',
]



#!/usr/bin/env python
import sys
from django.core.management import execute_manager
try:
    from . import settings  # Assumed to be in the same directory.
except ImportError:
    sys.stderr.write(
        "Error: Can't find the file 'settings.py' in the directory containing %r."
        " It appears you've customized things.\n"
        "You'll have to run django-admin.py, passing it your settings module.\n"
        "(If the file settings.py does indeed exist, it's causing an ImportError somehow.)\n"
        % __file__
    )
    sys.exit(1)

sys.argv.insert(1, 'test')

if len(sys.argv) == 2:
    sys.argv.append('payfast')

if __name__ == "__main__":
    execute_manager(settings)



# coding: utf-8
from __future__ import unicode_literals

import json
import unittest
from collections import OrderedDict

import django
from django.conf import settings
from django.contrib.auth.models import User
from django.test import TestCase, SimpleTestCase, override_settings

from payfast import api
from payfast import conf
from payfast.forms import notify_url, PayFastForm, is_payfast_ip_address
from payfast.models import PayFastOrder
import payfast.signals


class PayFastFormTest(TestCase):

    def test_init(self):
        form = PayFastForm(initial={
            'amount': 100,
            'item_name': 'Example item',
        })
        self.assertEqual({
            'amount': 100,
            'item_name': 'Example item',
            'm_payment_id': '1',
            'merchant_id': '10000100',
            'merchant_key': '46f0cd694581a',
            'notify_url': notify_url(),
        }, form.initial)
        self.assertIsNone(form.order.user)

    def test_init_with_user(self):
        user = User.objects.create(
            username='example_user',
            email='user@example.com',
            first_name='First',
            last_name='Last',
        )
        form = PayFastForm(
            initial={
                'amount': 100,
                'item_name': 'Example item',
            },
            user=user
        )
        self.assertEqual({
            'amount': 100,
            'email_address': 'user@example.com',
            'item_name': 'Example item',
            'm_payment_id': '1',
            'merchant_id': '10000100',
            'merchant_key': '46f0cd694581a',
            'name_first': 'First',
            'name_last': 'Last',
            'notify_url': notify_url(),
        }, form.initial)
        self.assertEqual(user, form.order.user)

    @override_settings(
        PAYFAST_GET_USER_FIRST_NAME=lambda user: user.username + ' First Name',
        PAYFAST_GET_USER_LAST_NAME=lambda user: user.username + ' Last Name',
    )
    def test_init_with_user_custom_names(self):
        user = User.objects.create(
            username='example_user',
            email='user@example.com',
            first_name='First',
            last_name='Last',
        )
        form = PayFastForm(
            initial={
                'amount': 100,
                'item_name': 'Example item',
            },
            user=user
        )
        self.assertEqual({
            'amount': 100,
            'email_address': 'user@example.com',
            'item_name': 'Example item',
            'm_payment_id': '1',
            'merchant_id': '10000100',
            'merchant_key': '46f0cd694581a',
            'name_first': 'example_user First Name',
            'name_last': 'example_user Last Name',
            'notify_url': notify_url(),
        }, form.initial)
        self.assertEqual(user, form.order.user)

    @override_settings(
        PAYFAST_GET_USER_FIRST_NAME=None,
        PAYFAST_GET_USER_LAST_NAME=None,
    )
    def test_init_with_user_custom_names_disabled(self):
        user = User.objects.create(
            username='example_user',
            email='user@example.com',
            first_name='First',
            last_name='Last',
        )
        form = PayFastForm(
            initial={
                'amount': 100,
                'item_name': 'Example item',
            },
            user=user
        )
        self.assertEqual({
            'amount': 100,
            'email_address': 'user@example.com',
            'item_name': 'Example item',
            'm_payment_id': '1',
            'merchant_id': '10000100',
            'merchant_key': '46f0cd694581a',
            'notify_url': notify_url(),
        }, form.initial)
        self.assertEqual(user, form.order.user)


def _test_data():
    return OrderedDict([
        ('merchant_id', '10000100'),
        ('merchant_key', '46f0cd694581a'),
        ('notify_url', "http://127.0.0.1:8000/payfast/notify/"),
        ('name_first', "Вася"),
        ('name_last', 'Пупников'),
        ('m_payment_id', '23'),
        ('amount', '234'),
        ('item_name', "Payment (Планета суши). ID:272-15"),
    ])


def _itn_data_from_checkout(checkout_data, payment_form):
    notify_data = checkout_data.copy()
    # prepare server data
    notify_data['m_payment_id'] = payment_form.order.m_payment_id
    notify_data['amount_gross'] = checkout_data['amount']
    # Fields not part of ITN submissions:
    del notify_data['notify_url']
    del notify_data['amount']
    del notify_data['merchant_key']

    # Sign:
    notify_data['signature'] = api.itn_signature(notify_data)
    return notify_data


def _order():
    return PayFastOrder.objects.all()[0]


class SignatureTest(unittest.TestCase):
    # TODO: This needs better coverage.

    def test_checkout_signature(self):
        data = _test_data()
        self.assertEqual(api.checkout_signature(data), '481366608545707be67c6514386b3fb1')

    def test_checkout_signature_blank_fields(self):
        """
        Fields with blank values should not be included in the signature.
        """
        data = _test_data()
        data['name_first'] = ''
        self.assertEqual(api.checkout_signature(data), '6551205f0fee13cf09174b0b887ec5b3')
        data['name_last'] = ''
        self.assertEqual(api.checkout_signature(data), '8f6435965cd9b00a9a965d93fc6c4c48')

    def test_known_good_itn_signature(self):
        known_good_itn_data = {
            'amount_fee': '-2.80',
            'amount_gross': '123.00',
            'amount_net': '120.20',
            'custom_int1': '',
            'custom_int2': '',
            'custom_int3': '',
            'custom_int4': '',
            'custom_int5': '',
            'custom_str1': '',
            'custom_str2': '',
            'custom_str3': '',
            'custom_str4': '',
            'custom_str5': '',
            'email_address': 'sbtu01@payfast.co.za',
            'item_description': '',
            'item_name': 'Flux capacitor',
            'm_payment_id': '',
            'merchant_id': '10000100',
            'name_first': 'Test',
            'name_last': 'User 01',
            'payment_status': 'COMPLETE',
            'pf_payment_id': '558900',
            'signature': '94b05677771813701468289ed3cabed1',
        }

        calculated_signature = api.itn_signature(known_good_itn_data)
        assert known_good_itn_data['signature'] == calculated_signature


@override_settings(PAYFAST_IP_ADDRESSES=['127.0.0.1'])
class NotifyTest(TestCase):

    def setUp(self):
        conf.USE_POSTBACK = False
        conf.MERCHANT_ID = '10000100'
        conf.REQUIRE_AMOUNT_MATCH = True

        self.notify_handler_orders = []  # type: list
        payfast.signals.notify.connect(self.notify_handler)

    def tearDown(self):
        payfast.signals.notify.disconnect(self.notify_handler)

    def notify_handler(self, sender, order, **kwargs):
        self.notify_handler_orders.append(order)

    def _create_order(self):
        """
        Create a payment order, and return the notification data for it.
        """
        checkout_data = _test_data()

        # user posts the pay request
        payment_form = PayFastForm(initial={
            'amount': checkout_data['amount'],
            'item_name': checkout_data['item_name']
        })
        self.assertEqual(_order().trusted, None)

        return _itn_data_from_checkout(checkout_data, payment_form)

    def _assertBadRequest(self, response, expected_json):
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response['Content-Type'], 'application/json')
        response_json = json.loads(response.content)
        self.assertEqual(response_json, expected_json)

    def test_notify(self):
        notify_data = self._create_order()
        order = _order()

        # the server sends a notification
        response = self.client.post(notify_url(), notify_data)
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(self.notify_handler_orders, [order])

        order = _order()
        self.assertEqual(order.request_ip, '127.0.0.1')
        self.assertEqual(order.debug_info, '')
        self.assertEqual(order.trusted, True)

    def test_untrusted_ip(self):
        """
        The notify handler rejects notification attempts from untrusted IP address.
        """
        notify_data = self._create_order()

        # the server sends a notification
        response = self.client.post(notify_url(), notify_data, REMOTE_ADDR='127.0.0.2')
        self._assertBadRequest(response, {
            '__all__': [{'code': '', 'message': 'untrusted ip: 127.0.0.2'}],
        })
        self.assertEqual(self.notify_handler_orders, [])

        order = _order()
        self.assertEqual(order.request_ip, '127.0.0.2')
        self.assertEqual(order.debug_info, '__all__: untrusted ip: 127.0.0.2')
        self.assertEqual(order.trusted, False)

    def test_non_existing_order(self):
        response = self.client.post(notify_url(), {})
        self.assertEqual(response.status_code, 404)
        self.assertEqual(self.notify_handler_orders, [])

        self.assertQuerysetEqual(PayFastOrder.objects.all(), [])

    def test_invalid_request(self):
        form = PayFastForm(initial={'amount': 100, 'item_name': 'foo'})
        notify_data = {'m_payment_id': form.order.m_payment_id}
        notify_data['signature'] = api.itn_signature(notify_data)
        response = self.client.post(notify_url(), notify_data)
        expected_amount = ('100' if django.VERSION < (1, 8) else
                           '100.00')
        self._assertBadRequest(response, {
            'amount_gross': [{'code': '', 'message': ('Amount is not the same: {} != None'
                                                      .format(expected_amount))}],
            'item_name': [{'code': 'required', 'message': 'This field is required.'}],
            'merchant_id': [{'code': 'required', 'message': 'This field is required.'}],
        })
        self.assertEqual(self.notify_handler_orders, [])

        order = _order()
        self.assertEqual(order.request_ip, '127.0.0.1')
        self.assertEqual(set(order.debug_info.split('|')), {
            'amount_gross: Amount is not the same: {} != None'.format(
                '100' if django.VERSION < (1, 8) else
                # Django 1.8+ returns more precise DecimalField values
                '100.00'
            ),
            'item_name: This field is required.',
            'merchant_id: This field is required.',
        })
        self.assertEqual(order.trusted, False)


class IPTest(SimpleTestCase):

    @override_settings(PAYFAST_IP_ADDRESSES=[])
    def test_no_addresses(self):
        self.assertFalse(is_payfast_ip_address('127.0.0.1'))
        self.assertFalse(is_payfast_ip_address('41.74.179.194'))

    @override_settings(PAYFAST_IP_ADDRESSES=['127.0.0.1'])
    def test_localhost(self):
        self.assertTrue(is_payfast_ip_address('127.0.0.1'))
        self.assertFalse(is_payfast_ip_address('41.74.179.194'))

    @override_settings(PAYFAST_IP_ADDRESSES=['41.74.179.194'])
    def test_one_server(self):
        self.assertFalse(is_payfast_ip_address('127.0.0.1'))
        self.assertTrue(is_payfast_ip_address('41.74.179.194'))

    @override_settings(PAYFAST_IP_ADDRESSES=['196.33.227.224', '196.33.227.225'])
    def test_more_servers(self):
        self.assertFalse(is_payfast_ip_address('127.0.0.1'))
        self.assertFalse(is_payfast_ip_address('41.74.179.194'))
        self.assertTrue(is_payfast_ip_address('196.33.227.224'))
        self.assertTrue(is_payfast_ip_address('196.33.227.225'))

    @override_settings(PAYFAST_IP_ADDRESSES=['196.33.227.224/31'])
    def test_more_servers_masked(self):
        self.assertFalse(is_payfast_ip_address('127.0.0.1'))
        self.assertFalse(is_payfast_ip_address('41.74.179.194'))

        self.assertFalse(is_payfast_ip_address('196.33.227.223'))
        self.assertTrue(is_payfast_ip_address('196.33.227.224'))
        self.assertTrue(is_payfast_ip_address('196.33.227.225'))
        self.assertFalse(is_payfast_ip_address('196.33.227.226'))

    @override_settings(PAYFAST_IP_ADDRESSES=[])
    def test_default_ip_addresses(self):
        del settings.PAYFAST_IP_ADDRESSES

        self.assertFalse(is_payfast_ip_address('127.0.0.1'))
        self.assertFalse(is_payfast_ip_address('196.33.227.224'))

        # Default PayFast range: 197.97.145.144/28
        self.assertFalse(is_payfast_ip_address('197.97.145.143'))
        self.assertTrue(all(is_payfast_ip_address('197.97.145.{}'.format(n))
                            for n in range(144, 160)))
        self.assertFalse(is_payfast_ip_address('197.97.145.160'))

        # Default PayFast range: 41.74.179.192/27
        self.assertFalse(is_payfast_ip_address('41.74.179.191'))
        self.assertTrue(all(is_payfast_ip_address('41.74.179.{}'.format(n))
                            for n in range(192, 224)))
        self.assertFalse(is_payfast_ip_address('41.74.179.225'))



"""
Test helpers: ElementTree parsing.
"""
from typing import Union, Optional  # noqa: F401
from xml.etree.ElementTree import Element, ElementTree


def text_collapsed(element):  # type: (Element) -> str
    """
    Element's text with whitespace collapsed.
    """
    chunks = (s.strip() for s in element.itertext())
    return ' '.join(s for s in chunks if s)


def text_lines(element):  # type: (Element) -> str
    """
    Element's text with lines preserved.
    """
    raw = ''.join(element.itertext())
    # We're probably not interested in outer empty lines.
    outer_stripped = raw.strip()
    # Strip each line.
    # Append newlines on all lines, including the last,
    # to ease comparison with dedent().
    stripped_linewise = ''.join(
        line.strip() + '\n'
        for line in outer_stripped.splitlines()
    )
    return stripped_linewise


_Element = Union[ElementTree, Element]


def find_id_maybe(base, id):  # type: (_Element, str) -> Optional[Element]
    """
    Return the child element with the given id, or None.
    """
    assert id.replace('-', '').isalnum(), id
    matches = base.findall('.//*[@id="{}"]'.format(id))
    if 0 == len(matches):
        return None
    elif 1 == len(matches):
        [match] = matches
        return match
    else:
        raise ValueError(
            'Found {} matching elements with id {!r}'.format(len(matches), id),
            matches,
        )


def find_id(base, id):  # type: (_Element, str) -> Element
    """
    Return the child element with the given id.

    :raise ValueError: If not found.
    """
    match = find_id_maybe(base, id)
    if match is None:
        raise ValueError('Element with id {!r} not found'.format(id))
    else:
        return match



"""
Integration tests against the PayFast sandbox environment.
"""
from __future__ import unicode_literals

import decimal
import os
from queue import Queue  # noqa: F401
from textwrap import dedent
from typing import Dict, Iterable, Tuple, Mapping  # noqa: F401
from xml.etree.ElementTree import ElementTree  # noqa: F401

from payfast import api

import html5lib
import requests

import pytest

from etree_helpers import text_collapsed, text_lines, find_id_maybe, find_id
from itn_helpers import itn_handler


# PayFast sandbox details
sandbox_process_url = 'https://sandbox.payfast.co.za/eng/process'
sandbox_merchant_credentials = {
    'merchant_id': '10000100',
    'merchant_key': '46f0cd694581a',
}


def post_sandbox_checkout(checkout_data):  # type: (Dict[str, str]) -> requests.Response
    """
    Post data to the PayFast sandbox checkout process endpoint.
    """
    response = requests.post(sandbox_process_url, checkout_data)
    response.raise_for_status()
    return response


def post_sandbox_payment(
        session_type,  # type: str
        session_id,  # type: str
        selected_method,  # type: str
):  # type: (...) -> requests.Response
    """
    Post a PayFast sandbox wallet payment confirmation.

    The parameters should come from the checkout page.
    """
    # This call is referenced from:
    # https://sandbox.payfast.co.za/js/engine_v2.js?version=5.2.6
    # (See the #pay-with-wallet click handler.)
    url = sandbox_process_url + '/payment_method?{}={}'.format(session_type, session_id)
    response = requests.post(url, {'selected_method': selected_method})
    response.raise_for_status()
    return response


def parse_payfast_page(response):  # type: (requests.Response) -> Dict[str, str]
    """
    Scrape some data from a PayFast payment page response.
    """
    assert 'text/html; charset=UTF-8' == response.headers['Content-Type']
    html = response.text
    doc = html5lib.parse(html)  # type: ElementTree

    def _parse():  # type: () -> Iterable[Tuple[str, str]]
        # The session info:
        session_tracker = find_id(doc, 'session-tracker')
        for name in ['type', 'id']:
            value = session_tracker.attrib['data-{}'.format(name)]
            if value:
                yield ('session_{}'.format(name), value)

        # The payment summary on the left.
        left = find_id(doc, 'left-column')
        yield ('payment_summary', text_collapsed(left))

        right = find_id(doc, 'right-column')
        content_box = find_id(right, 'contentBox')

        # The error notice, if any:
        notice = find_id_maybe(content_box, 'notice')
        if notice is not None:
            yield ('notice', text_lines(notice))

        # The wallet payment completion option, if present:
        wa_tab = find_id_maybe(content_box, 'waTab')
        if wa_tab is not None:
            yield ('payment_method', (wa_tab.attrib['data-methodkey']))
            pay_button = find_id(wa_tab, 'pay-with-wallet')
            yield ('pay_button', pay_button.attrib['value'])

    return dict(_parse())


def test_process_empty():  # type: () -> None
    """
    Submitting an empty payment request fails.
    """
    response = post_sandbox_checkout({})
    assert {
        'payment_summary': 'Payment total R ZAR',
        'notice': dedent("""\
            The supplied variables are not according to specification:
            amount : amount is required
            item_name : item_name is required
            merchant_id : merchant_id is required
            merchant_key : merchant_key is required
        """),
    } == parse_payfast_page(response)


# Check for ITN testing configuration:
try:
    ITN_HOST = os.environ['ITN_HOST']
    ITN_PORT = int(os.environ['ITN_PORT'])
    ITN_URL = os.environ['ITN_URL']
except KeyError:
    itn_configured = False
else:
    itn_configured = True


ITN_HELP_MESSAGE = """\
Configure the following environment variables to test ITN:
    ITN_HOST: The local host to listen on
    ITN_PORT: The local port to listen on
    ITN_URL: The notify_url to pass to PayFast
"""

#: Helper pytest mark for tests than need ITN handling configured.
requires_itn_configured = pytest.mark.skipif(not itn_configured, reason=ITN_HELP_MESSAGE)


def require_itn_configured():  # type: () -> None
    """
    pytest.skip() if not itn_configured.
    """
    if not itn_configured:
        pytest.skip(ITN_HELP_MESSAGE)


def do_checkout(
        checkout_data,  # type: Dict[str, str]
        sign_checkout,  # type: bool
):  # type: (...) -> Dict[str, str]
    """
    Common test helper: do a checkout, and assert results.

    This takes unsigned checkout data, and will add a signature if `sign_checkout` is true.

    Return the checkout page's parse.
    """
    # Expected values for result assertions:
    try:
        expected_amount = '{:.2f}'.format(decimal.Decimal(checkout_data['amount']))
    except decimal.InvalidOperation:
        # We may be testing a value that isn't Decimal-parseable;
        # in that case, just expect it unmodified.
        expected_amount = checkout_data['amount']
    expected_item_name = checkout_data['item_name'].strip()  # PayFast strips this for display.
    expected_payment_summary = (
        '{} Payment total R {} ZAR'.format(expected_item_name, expected_amount)
        .strip()  # Strip to handle item names that render empty.
    )

    if sign_checkout:
        assert 'signature' not in checkout_data, checkout_data
        checkout_data['signature'] = api.checkout_signature(checkout_data)

    response = post_sandbox_checkout(checkout_data)
    parsed = parse_payfast_page(response)
    assert {
        'session_type': 'p-sb',
        'session_id': parsed.get('session_id', 'MISSING'),
        'payment_summary': expected_payment_summary,
        'payment_method': '1',
        'pay_button': 'Complete Payment',
    } == parsed

    return parsed


def do_payment(
        checkout_data,  # Dict[str, str]
        parsed_checkout,  # Dict[str, str]
        enable_itn,  # type: bool
):  # type: (...) -> Dict[str, str]
    """
    Common test helper: do a payment, and assert results.

    This takes a checkout's data and page parse (for session info and assertions).
    This will enable and verify ITN processing if `enable_itn` is true.

    Return the payment confirmation page's parse.
    """
    def _post_payment():  # type: () -> requests.Response
        return post_sandbox_payment(
            parsed_checkout['session_type'],
            parsed_checkout['session_id'],
            parsed_checkout['payment_method'],
        )

    if enable_itn:
        require_itn_configured()
        with itn_handler(ITN_HOST, ITN_PORT) as itn_queue:  # type: Queue
            response = _post_payment()
            itn_data = itn_queue.get(timeout=2)
    else:
        response = _post_payment()

    parsed_payment = parse_payfast_page(response)
    assert {
        'payment_summary': parsed_checkout['payment_summary'],
        'notice': 'Your payment was successful\n'
    } == parsed_payment

    if enable_itn:
        # Check the ITN result.

        # Expect whitespace-stripped versions of the checkout data.
        expected = {name: value.strip(api.CHECKOUT_SIGNATURE_IGNORED_WHITESPACE)
                    for (name, value) in checkout_data.items()}

        expected_amount_gross = '{:.2f}'.format(decimal.Decimal(checkout_data['amount'].strip()))

        expected_signature = api.itn_signature(itn_data)
        assert {
            'm_payment_id': expected.get('m_payment_id', ''),
            'pf_payment_id': itn_data.get('pf_payment_id', 'MISSING'),
            'payment_status': 'COMPLETE',
            'item_name': expected.get('item_name', 'MISSING'),
            'item_description': expected.get('item_description', ''),

            'amount_gross': expected_amount_gross,
            'amount_fee': itn_data.get('amount_fee', 'MISSING'),
            'amount_net': itn_data.get('amount_net', 'MISSING'),

            'custom_str1': expected.get('custom_str1', ''),
            'custom_str2': expected.get('custom_str2', ''),
            'custom_str3': expected.get('custom_str3', ''),
            'custom_str4': expected.get('custom_str4', ''),
            'custom_str5': expected.get('custom_str5', ''),
            'custom_int1': expected.get('custom_int1', ''),
            'custom_int2': expected.get('custom_int2', ''),
            'custom_int3': expected.get('custom_int3', ''),
            'custom_int4': expected.get('custom_int4', ''),
            'custom_int5': expected.get('custom_int5', ''),

            # The sandbox seems to fix these names, rather than using the checkout submission data.
            'name_first': 'Test',
            'name_last': 'User 01',

            'email_address': expected.get('email_address', 'sbtu01@payfast.co.za'),
            'merchant_id': '10000100',
            'signature': expected_signature,
        } == itn_data

    return parsed_payment


def do_complete_payment(
        partial_checkout_data,  # type: Mapping[str, str]
        sign_checkout,  # type: bool
        enable_itn,  # type: bool
):  # type: (...) -> None
    """
    Common test helper: A complete checkout + payment completion flow.

    This takes a partial checkout data, and will add sandbox merchant credentials,

     * Sandbox merchant credentials (if not supplied)
     * notify_url (if enable_itn is true)
    """
    if enable_itn:
        require_itn_configured()

    # Step 1: Prepare the checkout data:
    checkout_data = {}
    checkout_data.update(sandbox_merchant_credentials)
    checkout_data.update(partial_checkout_data)

    if enable_itn:
        assert 'notify_url' not in checkout_data, checkout_data
        checkout_data['notify_url'] = ITN_URL

    # Step 2: Checkout process.
    parsed_checkout = do_checkout(checkout_data, sign_checkout=sign_checkout)

    # Step 3: Complete payment.
    do_payment(checkout_data, parsed_checkout, enable_itn=enable_itn)


def test_minimal_successful_payment_unsigned():  # type: () -> None
    """
    A minimal process + payment flow.
    """
    checkout_data = {
        'amount': '123',
        'item_name': 'Flux capacitor',
    }
    do_complete_payment(checkout_data, sign_checkout=False, enable_itn=False)


def test_minimal_successful_payment_signed():  # type: () -> None
    """
    A minimal process + payment flow, with signature.
    """
    checkout_data = {
        'amount': '123',
        'item_name': 'Flux capacitor',
    }
    do_complete_payment(checkout_data, sign_checkout=True, enable_itn=False)


@requires_itn_configured
def test_minimal_payment_itn():  # type: () -> None
    """
    A minimal payment with ITN.
    """
    checkout_data = {
        'amount': '123',
        'item_name': 'Flux capacitor',
    }
    do_complete_payment(checkout_data, sign_checkout=True, enable_itn=True)


@pytest.mark.parametrize(('leading', 'trailing'), [

    # These whitespace characters should be ignored:
    ('', ' '),
    (' ', ''),
    (' ', ' '),
    ('\t', '\t'),
    ('\n', '\n'),
    ('\r', '\r'),
    ('\x0b', '\x0b'),  # \N{LINE TABULATION} (Python 2 does not know this Unicode character name)

    # XXX: PayFast seems to refuse null-containing values, so we can't test their signatures.
    # ('\0', '\0'),

    # These whitespace characters should be included:
    ('\N{NO-BREAK SPACE}', '\N{NO-BREAK SPACE}'),  # '\xa0'

])
def test_checkout_signature_ignored_whitespace(leading, trailing):  # type: (str, str) -> None
    """
    Checkout signatures should ignore certain leading and trailing whitespace in values.
    """
    checkout_data = {
        'merchant_id': '10000100',
        'merchant_key': '46f0cd694581a',
        'amount': '123',
        'item_name': 'Flux capacitor',
        'item_description': '1.21 jigowatts',
        'name_first': 'Emmet',
        'name_last': 'Brown',
        # TODO: Cover other field values?
    }
    whitespaced_data = {
        name: (leading + value + trailing)
        for (name, value) in checkout_data.items()
    }

    # Special-case handling: these three fields accept other whitespace, but not NBSP.
    # Just revert them for testing (but let the other fields still contain NBSP).
    for name in ['merchant_id', 'merchant_key', 'amount']:
        whitespaced_data[name] = whitespaced_data[name].strip('\N{NO-BREAK SPACE}')

    do_complete_payment(whitespaced_data, sign_checkout=True, enable_itn=True)


@requires_itn_configured
def test_item_name_nbsp():  # type: () -> None
    """
    NBSP is a valid item name.
    """
    checkout_data = {
        'amount': '123',
        'item_name': '\N{NO-BREAK SPACE}',
    }
    do_complete_payment(checkout_data, sign_checkout=True, enable_itn=True)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.utils import timezone
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='PayFastOrder',
            fields=[
                ('m_payment_id', models.AutoField(help_text="Unique transaction ID on the receiver's system.", serialize=False, primary_key=True)),
                ('pf_payment_id', models.CharField(help_text='Unique transaction ID on PayFast.', max_length=40, unique=True, null=True, blank=True)),
                ('payment_status', models.CharField(help_text='The status of the payment.', max_length=20, null=True, blank=True)),
                ('item_name', models.CharField(help_text='The name of the item being charged for.', max_length=100)),
                ('item_description', models.CharField(help_text='The description of the item being charged for.', max_length=255, null=True, blank=True)),
                ('amount_gross', models.DecimalField(help_text='The total amount which the payer paid.', null=True, max_digits=15, decimal_places=2, blank=True)),
                ('amount_fee', models.DecimalField(help_text='The total in fees which was deducated from the amount.', null=True, max_digits=15, decimal_places=2, blank=True)),
                ('amount_net', models.DecimalField(help_text="The net amount credited to the receiver's account.", null=True, max_digits=15, decimal_places=2, blank=True)),
                ('custom_str1', models.CharField(max_length=255, null=True, blank=True)),
                ('custom_str2', models.CharField(max_length=255, null=True, blank=True)),
                ('custom_str3', models.CharField(max_length=255, null=True, blank=True)),
                ('custom_str4', models.CharField(max_length=255, null=True, blank=True)),
                ('custom_str5', models.CharField(max_length=255, null=True, blank=True)),
                ('custom_int1', models.IntegerField(null=True, blank=True)),
                ('custom_int2', models.IntegerField(null=True, blank=True)),
                ('custom_int3', models.IntegerField(null=True, blank=True)),
                ('custom_int4', models.IntegerField(null=True, blank=True)),
                ('custom_int5', models.IntegerField(null=True, blank=True)),
                ('name_first', models.CharField(help_text='First name of the payer.', max_length=100, null=True, blank=True)),
                ('name_last', models.CharField(help_text='Last name of the payer.', max_length=100, null=True, blank=True)),
                ('email_address', models.CharField(help_text='Email address of the payer.', max_length=100, null=True, blank=True)),
                ('merchant_id', models.CharField(help_text='The Merchant ID as given by the PayFast system.', max_length=15)),
                ('signature', models.CharField(help_text='A security signature of the transmitted data', max_length=32, null=True, blank=True)),
                ('created_at', models.DateTimeField(default=timezone.now)),
                ('updated_at', models.DateTimeField(default=timezone.now)),
                ('request_ip', models.GenericIPAddressField(null=True, blank=True)),
                ('debug_info', models.CharField(max_length=255, null=True, blank=True)),
                ('trusted', models.NullBooleanField(default=None)),
                ('user', models.ForeignKey(blank=True, to=settings.AUTH_USER_MODEL, null=True, on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name': 'PayFast order',
                'verbose_name_plural': 'PayFast orders',
            },
        ),
    ]



"""
Hypothesis-driven integration testing against the PayFast sandbox environment.

Handle and run with care!

This should probably only be run interactively, and not as part of the normal test suite.
"""
from __future__ import unicode_literals

import string
import time
from typing import Dict, Mapping, Callable, List  # noqa: F401

from hypothesis import given, assume, settings
from hypothesis import strategies as st

from test_integration_sandbox import (
    sandbox_merchant_credentials,
    requires_itn_configured,
    do_complete_payment,
)


# The minimum set of checkout fields that are required by PayFast.
required_checkout_fields = {
    'merchant_id',
    'merchant_key',
    'amount',
    'item_name',
}


def valid_text(
        min_size=None,  # type: int
        average_size=None,  # type: int
        max_size=None,  # type: int
):  # type: (...) -> st.S
    """
    Helper strategy for valid PayFast field value strings.
    """
    return (st.text(min_size=min_size, average_size=average_size, max_size=max_size)
            # PayFast seems to categorically refuse (with a low-level HTTP 403 Forbidden)
            # requests with values that contain nulls, so avoid generating these.
            .filter(lambda s: '\0' not in s))


@st.composite
def st_checkout_data(draw):  # type: (Callable) -> Mapping[str, str]
    """
    Strategy for generating valid checkout data.
    """
    # Requesting large amounts drains the sandbox wallet balance quickly, before it can be refilled.
    # When this happens, requesting amounts over the wallet balance
    # seems to result in a redirect loop, instead of a clean error.
    max_amount = 1000

    # Valid checkout amounts:
    st_amount = (st.decimals(min_value=0, max_value=max_amount,
                             allow_nan=False, allow_infinity=False)
                 .filter(lambda amount: amount != 0)  # The amount can't be exactly zero.
                 .map(str))

    # The required keys:
    checkout_data = dict(sandbox_merchant_credentials)
    checkout_data.update({
        'amount': draw(st_amount),
        'item_name': draw(valid_text(min_size=1, max_size=100)),
    })

    # The m_payment_id field has a specific limited character repertoire.
    m_payment_id_alphabet = string.ascii_letters + string.digits + '_-/'

    # Optional keys
    optional_fields_spec = {

        # Merchant Details
        # Already required: 'merchant_id',
        # Already required: 'merchant_key',
        # 'return_url',
        # 'cancel_url',
        # 'notify_url',

        # Buyer Detail
        'name_first': valid_text(max_size=100),
        'name_last': valid_text(max_size=100),
        # 'email_address',
        # 'cell_number',

        # Transaction Details
        'm_payment_id': st.text(alphabet=m_payment_id_alphabet, max_size=100),
        # Already required: 'amount',
        # Already required: 'item_name',
        'item_description': valid_text(max_size=255),
        'custom_int1': st.text(alphabet=string.digits, max_size=255),
        'custom_int2': st.text(alphabet=string.digits, max_size=255),
        'custom_int3': st.text(alphabet=string.digits, max_size=255),
        'custom_int4': st.text(alphabet=string.digits, max_size=255),
        'custom_int5': st.text(alphabet=string.digits, max_size=255),
        'custom_str1': valid_text(max_size=255),
        'custom_str2': valid_text(max_size=255),
        'custom_str3': valid_text(max_size=255),
        'custom_str4': valid_text(max_size=255),
        'custom_str5': valid_text(max_size=255),

        # Transaction Options
        # 'email_confirmation',
        # 'confirmation_address',

        # Set Payment Method
        # 'payment_method',

        # Recurring Billing Details
        # 'subscription_type',
        # 'billing_date',
        # 'recurring_amount',
        # 'frequency',
        # 'cycles',

    }

    # Make the fields optional by wrapping the values in singleton lists (simulating a Maybe type),
    # and flattening and skipping empty ones.
    optional_fields_intermediate = draw(st.fixed_dictionaries({
        k: st.lists(s, max_size=1)
        for (k, s) in optional_fields_spec.items()
    }))  # type: Dict[str, List[str]]
    # Flatten
    optional_fields = {
        k: v[0]
        for (k, v) in optional_fields_intermediate.items()
        if v
    }  # type: Dict[str, str]

    checkout_data.update(optional_fields)

    return checkout_data


@requires_itn_configured
@settings(
    # Attempt a vaguely API-friendly configuration:
    max_examples=10,
    max_iterations=1,
    max_shrinks=10,
    deadline=5000,
)
@given(st_checkout_data())
def test_complete_payment(checkout_data):  # type: (Mapping[str, str]) -> None
    assert required_checkout_fields <= checkout_data.keys()

    # XXX: PayFast treats '0' this as empty?
    assume(checkout_data['item_name'] != '0')

    # XXX: PayFast seems to normalise '\r' -> '\n' ?
    assume('\r' not in checkout_data['item_name'])

    try:
        print('XXX Starting payment...', checkout_data)
        do_complete_payment(checkout_data, sign_checkout=True, enable_itn=True)
        print('XXX Successful!')
    except AssertionError:
        print('XXX Got assertion failure.')
        raise
    except Exception as e:
        print('XXX Got error:', e)
        raise
    finally:
        # XXX: Hacky rate limiting: one second between requests.
        time.sleep(1)



# -*- coding: utf-8 -*-
"""
This migration promotes `m_payment_id` to a full alphanumeric field,
and lets PayFastOrder have a conventional primary key.
"""
# Generated by Django 1.11.8 on 2017-12-14 18:07
from __future__ import unicode_literals

import django
from django.db import migrations, models


if django.VERSION < (1, 8):  # Django 1.7 compatibility: RunPython.noop was added in 1.8.
    def noop(apps, schema_editor):
        return None


def preserve_old_m_payment_id_values(apps, schema_editor):
    """
    Preserve the old `m_payment_id` values into the new field.
    """
    PayFastOrder = apps.get_model('payfast', 'PayFastOrder')

    for order in PayFastOrder.objects.all():
        assert order.m_payment_id is None, order
        order.m_payment_id = str(order.pk)
        order.save()


class Migration(migrations.Migration):

    dependencies = [
        ('payfast', '0002_update_payfastorder_timestamp_fields'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='payfastorder',
            options={'verbose_name': 'PayFast order'},
        ),
        migrations.AlterField(
            model_name='payfastorder',
            name='amount_fee',
            field=models.DecimalField(blank=True, decimal_places=2, help_text='The total in fees which was deducted from the amount.', max_digits=15, null=True),
        ),

        # Convert the existing m_payment_id to a normal Django primary key.
        migrations.RenameField(
            model_name='payfastorder',
            old_name='m_payment_id',
            new_name='id',
        ),
        migrations.AlterField(
            model_name='payfastorder',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),

        # Re-introduce m_payment_id as an alphanumeric
        migrations.AddField(
            model_name='payfastorder',
            name='m_payment_id',
            field=models.CharField(blank=True, help_text="Unique transaction ID on the receiver's system.", max_length=100, null=True, unique=True),
        ),
        migrations.RunPython(
            preserve_old_m_payment_id_values,
            reverse_code=(noop if django.VERSION < (1, 8) else
                          migrations.RunPython.noop),
        ),
    ]



"""
This module can be used without django
"""
from __future__ import unicode_literals

import sys
from typing import Mapping, Tuple, Sequence  # noqa: F401

from hashlib import md5

# Python 2 compatibility:
from six import text_type as str
from six.moves.urllib.error import HTTPError
from six.moves.urllib.parse import urlencode
from six.moves.urllib.request import urlopen

from django.conf import settings


POSTBACK_URL = '/eng/query/validate'
POSTBACK_SERVER = 'https://www.payfast.co.za'


#: Field order for checkout process submission signatures.
#: See: https://developers.payfast.co.za/documentation/#checkout-page
checkout_signature_field_order = [

    # Merchant Details
    'merchant_id',
    'merchant_key',
    'return_url',
    'cancel_url',
    'notify_url',

    # Buyer Detail
    'name_first',
    'name_last',
    'email_address',
    'cell_number',

    # Transaction Details
    'm_payment_id',
    'amount',
    'item_name',
    'item_description',
    'custom_int1',
    'custom_int2',
    'custom_int3',
    'custom_int4',
    'custom_int5',
    'custom_str1',
    'custom_str2',
    'custom_str3',
    'custom_str4',
    'custom_str5',

    # Transaction Options
    'email_confirmation',
    'confirmation_address',

    # Set Payment Method
    'payment_method',

    # Recurring Billing Details
    'subscription_type',
    'billing_date',
    'recurring_amount',
    'frequency',
    'cycles',

]


#: Field order for ITN submission signatures.
#: https://developers.payfast.co.za/documentation/#notify-page-itn
itn_signature_field_order = [

    # Transaction details
    'm_payment_id',
    'pf_payment_id',
    'payment_status',
    'item_name',
    'item_description',
    'amount_gross',
    'amount_fee',
    'amount_net',
    'custom_str1',
    'custom_str2',
    'custom_str3',
    'custom_str4',
    'custom_str5',
    'custom_int1',
    'custom_int2',
    'custom_int3',
    'custom_int4',
    'custom_int5',

    # Buyer details
    'name_first',
    'name_last',
    'email_address',

    # Merchant details
    'merchant_id',

    # Recurring billing details
    'token',

]


#: A sequence of ordered (key, value) variables that can be signed for PayFast.
SignableFields = Sequence[Tuple[str, str]]


def _prepare_signable_fields(
        valid_field_order,  # type: Sequence[str]
        data_fields,  # type: Mapping[str, str]
):  # type: (...) -> SignableFields
    """
    Prepare PayFast submission variables for signing, using the given field order.

    :raise ValueError:
        If `data_fields` contains any unexpected field names not in `valid_field_order`.
    """
    present_fields = (set(data_fields.keys()) if sys.version_info < (3,) else
                      data_fields.keys())
    extra_fields = present_fields - set(valid_field_order)
    if extra_fields:
        raise ValueError('Data contains unexpected fields: {!r}'.format(extra_fields))

    return [
        (name, data_fields[name]) for name in valid_field_order
        if name in data_fields
    ]


def _drop_non_signature_fields(
        data_fields,  # type: Mapping[str, str]
        include_empty,  # type: bool
):  # type: (...) -> Mapping[str, str]
    """
    Drop fields that should not be included in signatures.

    These are:

        * `signature`, as a convenience for verifying already-signed data.
        * Fields with empty values, if `include_empty` is false.

    """
    return {
        k: v for (k, v) in data_fields.items()
        if k != 'signature'
        if include_empty or v
    }


def _sign_fields(signable_fields):  # type: (SignableFields) -> str
    """
    Common signing code.
    """
    for (k, v) in signable_fields:
        assert isinstance(k, str), repr(k)
        assert isinstance(v, str), repr(v)

    if sys.version_info < (3,):
        # Python 2 doesn't do IRI encoding.
        text = urlencode([
            (k.encode('utf-8'), v.encode('utf-8'))
            for (k, v) in signable_fields
        ])
    else:
        text = urlencode(signable_fields, encoding='utf-8', errors='strict')

    return md5(text.encode('ascii')).hexdigest()


#: The checkout signature should ignore these leading and trailing whitespace characters.
#:
#: This list is an educated guess based on the PHP trim() function.
#:
CHECKOUT_SIGNATURE_IGNORED_WHITESPACE = ''.join([
    ' ',
    '\t',
    '\n',
    '\r',
    '\x0b',  # \N{LINE TABULATION} (Python 2 does not know this Unicode character name)

    # XXX: trim() strips '\0', but it's not clear whether to actually strip it here.
    # We can't really test it, since the endpoint seems to refuse any requests with null values.
    # '\0',
])


def checkout_signature(checkout_data):  # type: (Mapping[str, str]) -> str
    """
    Calculate the signature of a checkout process submission.
    """
    # Omits fields with empty values.
    included_fields = _drop_non_signature_fields(checkout_data, include_empty=False)
    # Strip ignored whitespace from values.
    stripped_fields = {
        name: value.strip(CHECKOUT_SIGNATURE_IGNORED_WHITESPACE)
        for (name, value) in included_fields.items()
    }

    signable_fields = _prepare_signable_fields(checkout_signature_field_order, stripped_fields)
    return _sign_fields(signable_fields)


def itn_signature(itn_data):  # type: (Mapping[str, str]) -> str
    """
    Calculate the signature of an ITN submission.
    """
    # ITN signatures include fields with empty values.
    included_fields = _drop_non_signature_fields(itn_data, include_empty=True)
    signable_fields = _prepare_signable_fields(itn_signature_field_order, included_fields)
    return _sign_fields(signable_fields)


# TODO: Rework this and data_is_valid.
def _values_to_encode(data):
    return [
        (k, str(value).strip().encode('utf8'))
        for (k, value) in data.items()
        if k != 'signature'
    ]


def data_is_valid(post_data, postback_server=POSTBACK_SERVER):
    """
    Validates data via the postback. Returns True if data is valid,
    False if data is invalid and None if the request failed.
    """
    post_str = urlencode(_values_to_encode(post_data))  # type: str
    # FIXME: No Content-Type header.
    post_bytes = post_str.encode(settings.DEFAULT_CHARSET)  # type: bytes

    postback_url = postback_server.rstrip('/') + POSTBACK_URL
    try:
        response = urlopen(postback_url, data=post_bytes)
        result = response.read().decode('utf-8')  # XXX: Assumed encoding
    except HTTPError:
        # XXX: Just re-raise for now.
        raise

    if result == 'VALID':
        return True
    elif result == 'INVALID':
        return False
    else:
        raise NotImplementedError('Unexpected result from PayFast validation: {!r}'.format(result))



from __future__ import unicode_literals

import sys
from ipaddress import ip_address, ip_network
from operator import attrgetter

from django.contrib.auth import get_user_model
from six import text_type as str
from six.moves.urllib_parse import urljoin

import django
from django import forms
from django.conf import settings

from payfast import api
from payfast import conf
from payfast.models import PayFastOrder

# Django 1.10 introduces django.urls
if django.VERSION < (1, 10):
    from django.core.urlresolvers import reverse
else:
    from django.urls import reverse


def full_url(link):
    """
    Return an absolute version of a possibly-relative URL.

    This uses the PAYFAST_URL_BASE setting.
    """
    url_base = (conf.URL_BASE() if callable(conf.URL_BASE) else
                conf.URL_BASE)
    return urljoin(url_base, link)


def notify_url():
    return full_url(reverse('payfast_notify'))


class HiddenForm(forms.Form):
    """ A form with all fields hidden """
    def __init__(self, *args, **kwargs):
        super(HiddenForm, self).__init__(*args, **kwargs)
        for field in self.fields:
            self.fields[field].widget = forms.HiddenInput()


class PayFastForm(HiddenForm):
    """ PayFast helper form.
    It is not for validating data.
    It can be used to output html.

    Pass all the fields to form 'initial' argument. Form also has an optional
    'user' parameter: it is the User instance the order is purchased by. If
    'user' is specified, 'name_first', 'name_last' and 'email_address' fields
    will be filled automatically if they are not passed with 'initial'.

    If `m_payment_id` is specified, it will uniquely identify the PayFastOrder.
    Otherwise, a new PayFastOrder will be created for each form instantiation.
    """

    target = conf.PROCESS_URL

    # Receiver Details
    merchant_id = forms.CharField()
    merchant_key = forms.CharField()

    return_url = forms.URLField()
    cancel_url = forms.URLField()
    notify_url = forms.URLField()

    # Payer Details
    name_first = forms.CharField()
    name_last = forms.CharField()
    email_address = forms.CharField()
    # TODO: cell_number

    # Transaction Details
    m_payment_id = forms.CharField()
    amount = forms.CharField()
    item_name = forms.CharField()
    item_description = forms.CharField()

    custom_str1 = forms.CharField()
    custom_str2 = forms.CharField()
    custom_str3 = forms.CharField()
    custom_str4 = forms.CharField()
    custom_str5 = forms.CharField()

    custom_int1 = forms.IntegerField()
    custom_int2 = forms.IntegerField()
    custom_int3 = forms.IntegerField()
    custom_int4 = forms.IntegerField()
    custom_int5 = forms.IntegerField()

    # Transaction Options
    email_confirmation = forms.IntegerField()
    confirmation_address = forms.CharField()

    # Security
    signature = forms.CharField()

    def __init__(self, *args, **kwargs):
        get_first_name = getattr(settings, 'PAYFAST_GET_USER_FIRST_NAME', attrgetter('first_name'))
        get_last_name = getattr(settings, 'PAYFAST_GET_USER_LAST_NAME', attrgetter('last_name'))

        user = kwargs.pop('user', None)
        if user:

            if get_first_name is not None:
                kwargs['initial'].setdefault('name_first', get_first_name(user))
            if get_last_name is not None:
                kwargs['initial'].setdefault('name_last', get_last_name(user))

            # Django 1.11 adds AbstractBaseUser.get_email_field_name()
            email_address = (user.email if django.VERSION < (1, 11) else
                             getattr(user, get_user_model().get_email_field_name()))
            kwargs['initial'].setdefault('email_address', email_address)

        kwargs['initial'].setdefault('notify_url', notify_url())
        kwargs['initial'].setdefault('merchant_id', conf.MERCHANT_ID)
        kwargs['initial'].setdefault('merchant_key', conf.MERCHANT_KEY)

        super(PayFastForm, self).__init__(*args, **kwargs)

        if 'm_payment_id' in self.initial:
            # If the caller supplies m_payment_id, find the existing order, or create it.
            (self.order, created) = PayFastOrder.objects.get_or_create(
                m_payment_id=self.initial['m_payment_id'],
                defaults=dict(
                    user=user,
                    amount_gross=self.initial['amount'],
                ),
            )
            if not created:
                # If the order is existing, check the user and amount fields,
                # and update if necessary.
                #
                # XXX: Also consistency-check that the order is not paid yet?
                #
                if not (self.order.user == user and
                        self.order.amount_gross == self.initial['amount']):
                    self.order.user = user
                    self.order.amount_gross = self.initial['amount']
                    self.order.save()
        else:
            # Old path: Create a new PayFastOrder each time form is instantiated.
            self.order = PayFastOrder.objects.create(
                user=user,
                amount_gross=self.initial['amount'],
            )

            # Initialise m_payment_id from the pk.
            self.order.m_payment_id = str(self.order.pk)
            self.order.save()

            self.initial['m_payment_id'] = self.order.m_payment_id

        # Coerce values to strings, for signing.
        data = {k: str(v) for (k, v) in self.initial.items()}
        self._signature = self.fields['signature'].initial = api.checkout_signature(data)


def is_payfast_ip_address(ip_address_str):
    """
    Return True if ip_address_str matches one of PayFast's server IP addresses.

    Setting: `PAYFAST_IP_ADDRESSES`

    :type ip_address_str: str
    :rtype: bool
    """
    # TODO: Django system check for validity?
    payfast_ip_addresses = getattr(settings, 'PAYFAST_IP_ADDRESSES',
                                   conf.DEFAULT_PAYFAST_IP_ADDRESSES)

    if sys.version_info < (3,):
        # Python 2 usability: Coerce str to unicode, to avoid very common TypeErrors.
        # (On Python 3, this should generally not happen:
        #  let unexpected bytes values fail as expected.)
        ip_address_str = unicode(ip_address_str)  # noqa: F821
        payfast_ip_addresses = [unicode(address) for address in payfast_ip_addresses]  # noqa: F821

    return any(ip_address(ip_address_str) in ip_network(payfast_address)
               for payfast_address in payfast_ip_addresses)


class NotifyForm(forms.ModelForm):

    def __init__(self, request, *args, **kwargs):
        self.request = request
        super(NotifyForm, self).__init__(*args, **kwargs)
        # the form must be used with order instance provided
        assert self.instance.pk

    def clean(self):
        self.ip = self.request.META.get(conf.IP_HEADER, None)
        if not is_payfast_ip_address(self.ip):
            raise forms.ValidationError('untrusted ip: %s' % self.ip)

        # Verify signature
        sig = api.itn_signature(self.data)
        if sig != self.cleaned_data['signature']:
            raise forms.ValidationError('Signature is invalid: %s != %s' % (
                sig, self.cleaned_data['signature'],))

        if conf.USE_POSTBACK:
            is_valid = api.data_is_valid(self.request.POST, conf.SERVER)
            if is_valid is None:
                raise forms.ValidationError('Postback fails')
            if not is_valid:
                raise forms.ValidationError('Postback validation fails')

        return self.cleaned_data

    def clean_merchant_id(self):
        merchant_id = self.cleaned_data['merchant_id']
        if merchant_id != conf.MERCHANT_ID:
            raise forms.ValidationError('Invalid merchant id (%s).' % merchant_id)
        return merchant_id

    def clean_amount_gross(self):
        received = self.cleaned_data['amount_gross']
        if conf.REQUIRE_AMOUNT_MATCH:
            requested = self.instance.amount_gross
            if requested != received:
                raise forms.ValidationError('Amount is not the same: %s != %s' % (
                                            requested, received,))
        return received

    def save(self, *args, **kwargs):
        self.instance.request_ip = self.ip

        # Decode body, for saving as debug_info
        body_bytes = self.request.read()  # type: bytes
        body_encoding = (settings.DEFAULT_CHARSET if self.request.encoding is None else
                         self.request.encoding)
        body_str = body_bytes.decode(body_encoding)  # type: str

        self.instance.debug_info = body_str[:255]

        self.instance.trusted = True
        return super(NotifyForm, self).save(*args, **kwargs)

    def plain_errors(self):
        ''' plain error list (without the html) '''
        return '|'.join(["%s: %s" % (k, (v[0])) for k, v in self.errors.items()])

    class Meta:
        model = PayFastOrder
        exclude = ['created_at', 'updated_at', 'request_ip', 'debug_info',
                   'trusted', 'user']



# -*- coding: utf-8 -*-
# Generated by Django 1.11.8 on 2017-12-05 07:10
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('payfast', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='payfastorder',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True),
        ),
        migrations.AlterField(
            model_name='payfastorder',
            name='updated_at',
            field=models.DateTimeField(auto_now=True),
        ),
    ]



from __future__ import unicode_literals

import django
import os
import sys


# Turn timezone awareness violation warnings into errors, for development.
# See https://docs.djangoproject.com/en/stable/topics/i18n/timezones/#code
import warnings
warnings.filterwarnings(
    'error',
    r'DateTimeField .* received a naive datetime',
    RuntimeWarning,
    r'django\.db\.models\.fields',
)


PROJECT_ROOT = os.path.dirname(os.path.abspath(__file__))
join = lambda p: os.path.abspath(os.path.join(PROJECT_ROOT, p))  # noqa: E731

# use package from the folder above, not the installed version
sys.path.insert(0, join('..'))

# ===== payfast settings ====

PAYFAST_MERCHANT_ID = '10000100'
PAYFAST_MERCHANT_KEY = '46f0cd694581a'

PAYFAST_URL_BASE = 'http://example.com/'

# ===========================

DEBUG = True

ADMINS = ()
MANAGERS = ADMINS

DATABASES = {
    'default': {
        # Add 'postgresql_psycopg2', 'postgresql', 'mysql', 'sqlite3' or 'oracle'.
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': join('db.sqlite'),  # Or path to database file if using sqlite3.
    }
}

USE_TZ = True
TIME_ZONE = 'America/Chicago'
LANGUAGE_CODE = 'en-us'
USE_I18N = True
USE_L10N = True
MEDIA_ROOT = join('media')
MEDIA_URL = '/media/'
SECRET_KEY = '5mcs97ar-(nnxjok67290+0^sr!e(ax=x$2-!8dqy25ff-l1*a='

MIDDLEWARE = [
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
]

# Django 1.10 migrates MIDDLEWARE_CLASSES to MIDDLEWARE
if django.VERSION < (1, 10):
    MIDDLEWARE_CLASSES = MIDDLEWARE
    del MIDDLEWARE

ROOT_URLCONF = 'urls'

# Django 1.8 migrates TEMPLATE_* to TEMPLATES
if django.VERSION < (1, 8):
    TEMPLATE_DEBUG = DEBUG
    TEMPLATE_LOADERS = [
        'django.template.loaders.filesystem.Loader',
        'django.template.loaders.app_directories.Loader',
    ]
    TEMPLATE_DIRS = [
        join('templates'),
    ]
else:
    TEMPLATES = [
        {
            "BACKEND": 'django.template.backends.django.DjangoTemplates',
            'DIRS': (join('templates'),),
            'APP_DIRS': True,
            'OPTIONS': {
                'context_processors': [
                    'django.template.context_processors.debug',
                    'django.template.context_processors.request',
                    'django.contrib.auth.context_processors.auth',
                ],
            }
        }
    ]


INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.admin',
    'payfast',
]



import django
from django.conf.urls import url, include
from django.contrib import admin

import payfast.urls


if django.VERSION < (1, 7):
    admin.autodiscover()


urlpatterns = [
    url(r'^admin/', admin.site.urls),
    url(r'^payfast/', include(payfast.urls)),
]



#!/usr/bin/env python
import sys
from setuptools import setup, find_packages


def README():
    with open('README.rst') as f:
        return f.read()


_PYTHON_2_BACKPORTS = ['ipaddress'] if sys.version_info < (3,) else []


setup(
    name='django-payfast',
    # Version automatically from tags using setuptools-scm
    use_scm_version=True,

    maintainer='Pi Delport',
    maintainer_email='pjdelport@gmail.com',

    packages=find_packages(exclude=['payfast_tests']),

    setup_requires=['setuptools-scm'],

    install_requires=[
        'six',
        'Django',

        'typing; python_version<"3.5"',
    ] + _PYTHON_2_BACKPORTS,

    url='https://github.com/pjdelport/django-payfast',
    license='MIT license',
    description='A pluggable Django application for integrating payfast.co.za payment system.',
    long_description=README(),

    classifiers=(
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ),
)



from django.http import HttpResponse, HttpResponseBadRequest
from django.shortcuts import get_object_or_404
from django.views.decorators.csrf import csrf_exempt

from payfast.forms import NotifyForm
from payfast.models import PayFastOrder
from payfast import signals


@csrf_exempt
def notify_handler(request):
    """
    Notify URL handler.

    On successful access 'payfast.signals.notify' signal is sent.
    Orders should be processed in signal handler.
    """
    m_payment_id = request.POST.get('m_payment_id', None)
    order = get_object_or_404(PayFastOrder, m_payment_id=m_payment_id)

    form = NotifyForm(request, request.POST, instance=order)
    if not form.is_valid():
        errors = form.plain_errors()[:255]
        order.request_ip = form.ip
        order.debug_info = errors
        order.trusted = False
        order.save()

        # XXX: Any possible data leakage here?
        return HttpResponseBadRequest(
            content_type='application/json',
            content=form.errors.as_json(),
        )

    order = form.save()
    signals.notify.send(sender=notify_handler, order=order)
    return HttpResponse()



"""
Test helpers: Transient ITN handler.
"""
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from email.header import Header  # noqa: F401
from http import HTTPStatus
from http.server import HTTPServer, BaseHTTPRequestHandler
from queue import Queue
from typing import Dict, Iterator, Union  # noqa: F401
from urllib.parse import parse_qs


class PayFastITNHandler(BaseHTTPRequestHandler):
    """
    Minimal ITN callback handler, for testing.

    Expects the server to have a `itn_queue` to put the result to.
    """

    def do_POST(self):  # type: () -> None
        assert self.command == 'POST'
        assert self.path == '/'

        # Parse the request body, and post to the queue.
        data = self.read_request_data()
        itn_queue = self.server.itn_queue  # type: ignore
        itn_queue  # type: Queue
        itn_queue.put(data)

        self.send_response(HTTPStatus.OK)
        self.end_headers()

    def read_request_body(self):  # type: () -> str
        [content_length_header] = self.headers.get_all('Content-Length')  # type: Union[str, Header]
        content_length = int(str(content_length_header))
        content_bytes = self.rfile.read(content_length)
        content = content_bytes.decode('utf-8')  # XXX
        return content

    def read_request_data(self):  # type: () -> Dict[str, str]
        assert self.headers['Content-Type'] == 'application/x-www-form-urlencoded'

        content = self.read_request_body()
        data_lists = parse_qs(
            content,
            # Strict options:
            keep_blank_values=True,
            strict_parsing=True,
            errors='strict',
        )
        # Flatten the listed values.
        data_flat = {k: v for (k, [v]) in data_lists.items()}
        return data_flat


@contextmanager
def itn_handler(host, port):  # type: (str, int) -> Iterator[Queue]
    """
    Usage::

        with itn_handler(ITN_HOST, ITN_PORT) as itn_queue:
            # ...complete PayFast payment...
            itn_data = itn_queue.get(timeout=2)
    """
    server_address = (host, port)
    http_server = HTTPServer(server_address, PayFastITNHandler)
    http_server.itn_queue = Queue()  # type: ignore

    executor = ThreadPoolExecutor(max_workers=1)
    executor.submit(http_server.serve_forever)
    try:
        yield http_server.itn_queue  # type: ignore
    finally:
        http_server.shutdown()
